(* Word32.sml
 *
 * Peter Bertelsen
 * October 1997
 * modified by AW, 2004
 *)

structure Word32 :> WORD = struct 

open Word8Vector
val valOf = Option.valOf
val rev   = List.rev

datatype word = WORD of vector

val wordSize = 32

(* for WORD bs, bs[0] is the most significant byte *)

local
    open Word
    infix >> << orb
in
    fun fromWord w =
	let fun h 0w0 w' l = l
	      | h i   w' l =
	        h (i-0w1) (w' >> 0w8) (Word8.fromLargeWord w' :: l)
	in
	    WORD (fromList(h 0w4 w []))
	end

    fun toWord' (b, w) = (w << 0w8) orb Word8.toLargeWord b

    fun toWord (WORD bs) = foldl toWord' 0w0 bs
end

fun fromInt k =
    let fun h 0 k' l = l
	  | h i k' l = h (i-1) (k' div 0x100) (Word8.fromInt k' :: l)
    in
	WORD (fromList(h 4 k []))
    end

fun toInt' k0 bs =
    let fun h (_, b, k) = k * 0x100 + Word8.toInt b
    in
	foldli h k0 (bs, 1, NONE)
    end

fun toInt (WORD bs) =
    let val b0 = sub(bs, 0)
    in
	toInt' (Word8.toInt b0) bs
    end

fun toIntX (WORD bs) =
    let val b0 = sub(bs, 0)
    in
	toInt' (Word8.toIntX b0) bs
    end

fun fromBytes bs =
    if length bs = 4 then SOME (WORD bs)
    else NONE

fun toBytes (WORD bs) = bs

fun compare' x y i =
    let fun h 0 i' = Word8.compare(sub(x, i'), sub(y, i'))
	  | h k i' =
	    (case Word8.compare(sub(x, i'), sub(y, i')) of
		 EQUAL => h (k-1) (i'+1)
	       | res   => res)
    in
	h (3-i) i
    end

fun compare (WORD x, WORD y) = compare' x y 0
    
fun compareX (WORD x, WORD y) =
    let val x0 = sub(x, 0)
	val y0 = sub(y, 0)
    in
	if (Word8.andb(x0, 0wx80) = Word8.andb(y0, 0wx80)) then
	    (case Word8.compare(x0, y0) of
		 EQUAL => compare' x y 1
	       | res   => res)
	else Word8.compare(y0, x0)  (* sign(x) <> sign(y) *)
    end

fun emit out (WORD bs) = app out bs

(* FIXME: scan has a different interface now FIXME *)
(*
fun scan src =
    (let fun h 0 l = l
	   | h i l = h (i-1) (valOf(src()) :: l)

	 (* NOTE: src is imperative; we must use its results in the
	  * correct order: #[byte 0, byte 1, byte 2, and byte 3]. *)

	 val bs = fromList(rev(h 4 []))
     in
	 SOME (WORD bs)
     end) handle Option => NONE *)

val scan = raise Fail "the implementation of Word32.scan needs to be FIXED."

(* temporarily in MosmlCompat LargeWord = Word *)

val toLarge       = toWord
val toLargeX      = raise Fail "Word32.toLargeX FIXME."
val toLargeWord   = toLarge
val toLargeWordX  = toLargeX

val fromLarge     = fromWord
val fromLargeWord = fromLarge

(* LargeInt is Int in MosmlCompat *)

val toLargeInt   = toInt
val toLargeIntX  = toIntX
val fromLargeInt = fromInt

(* FIXME: More efficient? *)
fun mkbop ope (WORD a,WORD b) = WORD (tabulate(2, fn i => ope(sub(a,i),sub(b,i))))
fun mkuop ope (WORD a) = WORD (map ope a)

val andb = mkbop Word8.andb 
val orb  = mkbop Word8.orb
val xorb = mkbop Word8.xorb
val notb = mkuop Word8.notb

val <<         = raise Fail "<< Not implemented!"
val ~>>        = raise Fail "~<< Not implemented!"
val >>         = raise Fail ">> Not implemented!"
val ~>>        = raise Fail "~>> Not implemented!"
val op+        = raise Fail "+   Not implemented!"
val op-        = raise Fail "-   Not implemented!"
val op*        = raise Fail "*   Not implemented!"
val op div     = raise Fail "div Not implemented!"
val op mod     = raise Fail "mod Not implemented!"
val op <       = raise Fail "<  Not implemented!"
val op <=      = raise Fail "<= Not implemented!"
val op >       = raise Fail "<  Not implemented!"
val op >=      = raise Fail ">= Not implemented!"
val min        = raise Fail "min Not implemented!"
val max        = raise Fail "maz not implemented!"
val fmt        = raise Fail "fmt not implemented!"
val toString   = raise Fail "toString not implemented!"
val radix      = raise Fail "radix not implemented!"
val fromString = raise Fail "fromString not implemented!"

end